try:
    Text = basestring
except NameError:
    Text = (str, bytes)


class JobHandler(object):
    """

    """
    def __init__(self, datastorage, jobid):
        self._id = jobid
        self._data = datastorage
        self._inmemory_objects = dict()

    def __delitem__(self, key):
        if not isinstance(key, Text):
            raise ValueError("Field names must be strings")
        self._data.set_value(self._id, key, None, raw_parameter=True)

    def __getitem__(self, item):
        if not isinstance(item, Text):
            raise ValueError("Field names must be strings")

        try:
            return self._inmemory_objects[item]
        except KeyError:
            pass

        data = self._data.get_value(self._id, item)
        if data is None:
            raise KeyError("Field %s is NULL" % item)

        self._inmemory_objects[item] = data
        return data

    def __setitem__(self, key, value):
        if not isinstance(key, Text):
            raise ValueError("Field names must be strings")

        self._data.set_value(self._id, key, value)
        self._inmemory_objects[key] = value

    def __contains__(self, item):
        if not isinstance(item, Text):
            raise ValueError("Field names must be strings")

        try:
            value = self._inmemory_objects[item]
            return value is not None
        except KeyError:
            # No need to transform, just check if it is set
            return self._data.get_value(self._id, item, raw_return=True) is not None

    def get(self, item, default=None):
        if not isinstance(item, Text):
            raise ValueError("Field names must be strings")

        # Starts just like __getitem__
        try:
            return self._inmemory_objects[item]
        except KeyError:
            pass

        ret = self._data.get_value(self._id, item)

        if ret is None:
            return default
        else:
            return ret

    def setdefault(self, item, default=None):
        # does a get, and stores it **if** we are not receiving None
        ret = self.get(item, default)
        if ret is not None:
            self._inmemory_objects[item] = ret
        return ret

    def commit(self):
        self._data.set_values(self._id,
                              self._inmemory_objects.keys(),
                              self._inmemory_objects.values())
        self._data._conn.commit()

    def __del__(self):
        self.commit()



import json
try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    import yaml
except ImportError:
    yaml = None


class Identity(object):
    @staticmethod
    def from_db(data):
        return data

    @staticmethod
    def to_db(data):
        return data


class JsonTransformer(object):
    @staticmethod
    def from_db(data):
        return json.loads(bytes(data))

    @staticmethod
    def to_db(data):
        return json.dumps(data)


class YamlTransformer(object):
    @staticmethod
    def from_db(data):
        if yaml is None:
            raise ImportError("No YAML library available, YAML is unsupported")
        return yaml.load(bytes(data))

    @staticmethod
    def to_db(data):
        if yaml is None:
            raise ImportError("No YAML library available, YAML is unsupported")
        return yaml.dump(data)


class PickleTransformer(object):
    @staticmethod
    def from_db(data):
        return pickle.loads(bytes(data))

    @staticmethod
    def to_db(data):
        return pickle.dumps(data)



import os

from .dbconn import DataStorage, DummyDataStorage

class Slurm:
    JOB_ID = "SLURM_JOB_ID"
    TABLE_NAME = "SLURM_JOB_NAME"


class Pbs:
    JOB_ID = "PBS_JOBID"
    TABLE_NAME = "PBS_JOBNAME"


known_schedulers = [Slurm, Pbs]


def prepare_handler(scheduler_environ_vars):
    return DataStorage(
        os.environ[scheduler_environ_vars.TABLE_NAME]
    ).get_handler(int(os.environ[scheduler_environ_vars.JOB_ID]))


def prepare_dummy_handler():
    return DummyDataStorage().get_handler()


def auto_detect_table_name():
    for sch in known_schedulers:
        try:
            return os.environ[sch.TABLE_NAME]
        except KeyError:
            pass
    raise ValueError("Could not autodetect JOB_NAME (table name)")


def auto_detect_job_id():
    for sch in known_schedulers:
        try:
            return int(os.environ[sch.JOB_ID])
        except KeyError:
            pass
    raise ValueError("Could not autodetect JOB_ID (row identifier)")



from __future__ import print_function

import argparse
import json
import os
import sys
from datetime import datetime

try:
    import yaml
except ImportError:
    yaml = None

from . import schedulers
from .dbconn import DataStorage, DataSchema


def init(args):
    ds = DataSchema.load_from_file(args.schema)
    args.data_storage.prepare(ds)


def update(args):
    ds = DataSchema.load_from_file(args.schema)
    args.data_storage.update(ds)


def clear(args):
    args.data_storage.clear(remove_tables=args.remove_tables)


def _sanitize_job_id(args):
    if args.jobid:
        return int(args.jobid)
    else:
        return schedulers.auto_detect_job_id()


def get(args):
    job_id = _sanitize_job_id(args)
    value = args.data_storage.get_value(job_id, args.field, raw_return=True)
    print(value)


def set(args):
    job_id = _sanitize_job_id(args)

    if args.value == '-':
        value = sys.stdin.read()
    else:
        value = args.value
    args.data_storage.set_value(job_id, args.field, value, raw_parameter=True)


def setnow(args):
    job_id = _sanitize_job_id(args)
    args.data_storage.set_value(job_id, args.field, datetime.now())


def setdict(args):
    job_id = _sanitize_job_id(args)

    if args.file == '-':
        fp = sys.stdin
        if args.dialect is None:
            args.dialect = 'json'
    else:
        fp = open(args.file, 'rb')
        if args.dialect is None:
            if args.file.endswith(".json"):
                args.dialect = 'json'
            elif args.file.endswith(".yaml"):
                args.dialect = 'pickle'
            else:
                args.dialect = 'json'

    if args.dialect == 'json':
        d = json.load(fp)
    elif args.dialect == 'yaml':
        if yaml is None:
            raise ImportError("No YAML library available, cannot parse YAML documents")
        d = yaml.load(fp)

    fields, values = zip(*(d.items()))
    args.data_storage.set_values(job_id, fields, values, raw_parameters=True)


def main():
    parser = argparse.ArgumentParser("tad4bj")
    parser.add_argument('--database', '-d', action='store',
                        help='path to database file. Default: value of the environment '
                             'variable $TAD4BJ_DATABASE or, if absent, file ~/tad4bj.db'
                             ' || Current default value: %s'
                             % DataStorage.DATABASE_DEFAULT_PATH)
    parser.add_argument('--table', '-t', action='store',
                        help='name of the table to work with. If not present, tad4bj will'
                             'try to autodetect it from the jobname of the scheduler')
    parser.add_argument('--verbose', '-v', help='verbose output', action='count')

    subparsers = parser.add_subparsers(help='available commands')

    parser_init = subparsers.add_parser('init', help="Initialize a table in the database")
    parser_init.set_defaults(func=init)
    parser_init.add_argument('schema', action='store',
                             help='Schema file for the table that will be initialized')

    parser_init = subparsers.add_parser('update', help="Add fields to an already "
                                                       "existing table in the database")
    parser_init.set_defaults(func=update)
    parser_init.add_argument('schema', action='store',
                             help='Schema file for the table that will be initialized')

    parser_clear = subparsers.add_parser('clear')
    parser_clear.set_defaults(func=clear)
    parser_clear.add_argument('--remove-tables', '-r', action='store_true',
                              default=False,
                              help='Remove content and tables --requires `init` '
                                   'before the table can be used again')

    jobid_args = ('--jobid', '-j')
    jobid_kwargs = {
        'action': 'store',
        'help': 'Job identifier that will be used as row id. '
                'If not present, tad4bj will try to autodetect'
                'it from the jobid of the scheduler'
    }

    parser_get = subparsers.add_parser('get')
    parser_get.set_defaults(func=get)
    parser_get.add_argument(*jobid_args, **jobid_kwargs)
    parser_get.add_argument('field', action='store',
                            help='Name of the field that will be retrieved')

    parser_set = subparsers.add_parser('set')
    parser_set.set_defaults(func=set)
    parser_set.add_argument(*jobid_args, **jobid_kwargs)
    parser_set.add_argument('field', action='store',
                            help='Name of the field that will be set')
    parser_set.add_argument('value', action='store',
                            help='Value that will be assigned to the field')

    parser_setdict = subparsers.add_parser('setdict')
    parser_setdict.set_defaults(func=setdict)
    parser_setdict.add_argument(*jobid_args, **jobid_kwargs)
    parser_setdict.add_argument('--dialect', '-d', choices=['yaml', 'json'],
                                help='Serialization format of the dictionary used')
    parser_setdict.add_argument('file', action='store',
                                help='Path containing a file with a dictionary that will '
                                     'be used to assign values to fields')

    parser_setnow = subparsers.add_parser('setnow')
    parser_setnow.set_defaults(func=setnow)
    parser_setnow.add_argument(*jobid_args, **jobid_kwargs)
    parser_setnow.add_argument('field', action='store',
                               help='Name of the (timestamp-)field that will be set to "now"')

    args = parser.parse_args()

    if not args.table:
        args.table = schedulers.auto_detect_table_name()

    args.data_storage = DataStorage(args.table, path=args.database)

    args.func(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='tad4bj',
      version='1.1',
      description='Tabular Annotations of Data for Batch Jobs',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Alex Barcelo',
      author_email='alex@betarho.net',
      url='https://github.com/alexbarcelo/tad4bj',
      packages=['tad4bj'],
      extras_require={
          'yaml': ["pyyaml"],
      },
      entry_points={
          'console_scripts': {'tad4bj = tad4bj:main'},
      },
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: POSIX',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Topic :: Database',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: System :: Logging',
      ],
      )




import os
from collections import Mapping, namedtuple
import sqlite3
import json

try:
    import yaml
except ImportError:
    yaml = None

from . import transformers, handlers


class DataSchema(object):
    def __init__(self, dict):
        # ToDo: some error and sanity checking
        self.__dict__.update(dict)

    @classmethod
    def load_from_file(cls, path):
        """

        :param path:
        """
        with open(path, "r") as f:
            if path.endswith(".json"):
                dict_data = json.load(f)
            elif path.endswith(".yaml"):
                if not yaml:
                    raise ImportError("No YAML available --could not load %s file" % path)
                dict_data = yaml.load(f)
            else:
                raise NotImplementedError("File type not recognized, unable to read %s" % path)

        return cls(dict_data)


class DataStorage(Mapping):
    """Abstract the management of all the application data.

    This class will use a SQLite file and present it with more general
    interface to avoid having to cope with SQL and its internals.

    DISCLAIMER: I am using %s string formatting instead of the SQL intended ?
    because there where some issues in certain places (like table name) and
    the user has already access to the database itself, so there's no security
    issue.
    """
    DATABASE_DEFAULT_PATH = os.path.expanduser(os.getenv("TAD4BJ_DATABASE", "~/tad4bj.db"))

    def __init__(self, table_name, path=None):
        """
        :param table_name:
        :param path:
        """
        if path is None:
            path = DataStorage.DATABASE_DEFAULT_PATH

        self._conn = sqlite3.connect(
            path,
            detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES
        )
        self._cursor = self._conn.cursor()
        self._table = table_name
        self._metadata = None
        self._rowtuple = None
        self._field_transformers = dict()

    def _get_row_namedtuple(self):
        if self._rowtuple is not None:
            return self._rowtuple

        if not self._cursor.description:
            raise RuntimeError("Could not get row schema --empty cursor")

        self._rowtuple = namedtuple("Row%s" % self._table,
                                    (s[0] for s in self._cursor.description))
        return self._rowtuple

    def close(self):
        if self._conn:
            self._conn.commit()
            self._conn.close()
            self._conn = None

    def __del__(self):
        self.close()

    def to_dataframe(self):
        import pandas as pd

        df = pd.read_sql_query("SELECT * FROM %s" % self._table, self._conn)

        return df

    def clear(self, remove_tables=False):
        if remove_tables:
            # Drop them if they exist
            self._cursor.execute("DROP TABLE IF EXISTS `%s_tamd`" % self._table)
            self._cursor.execute("DROP TABLE IF EXISTS `%s`" % self._table)
        else:
            # Application should fail if the table does not exist, as this does:
            self._cursor.execute("DELETE FROM `%s`" % self._table)

    def prepare(self, schema):
        creation_fields = ", ".join("'%s' %s" % (field_name, field_type)
                                    for field_name, field_type in schema.fields)
        self._cursor.execute("CREATE TABLE `%s` (%s)" %
                             (self._table, creation_fields))
        self._metadata = schema.metadata
        self._cursor.execute("CREATE TABLE `%s_tamd` (field text, type text)" %
                             (self._table,))
        self._cursor.executemany("INSERT INTO `%s_tamd` (field, type) values (?, ?)" %
                                 (self._table,), self._metadata.items())
        # Exceptionally, the creation is forcefully committed
        self._conn.commit()

    def update(self, schema):
        self._cursor.execute("SELECT * FROM `%s`" % self._table)
        old_fields = {f[0] for f in self._cursor.description}
        all_fields = {f for f, _ in schema.fields}
        new_fields = all_fields - old_fields

        if not new_fields:
            return

        # Add all new columns
        map(self._cursor.execute,
            ("ALTER TABLE `%s` ADD COLUMN '%s' %s" %
             (self._table, field_name, field_type)
             for field_name, field_type in schema.fields
             if field_name in new_fields))

        self._metadata = schema.metadata

        # Add new elements into the metadata table
        self._cursor.executemany("INSERT INTO `%s_tamd` (field, type) values (?, ?)" %
                                 (self._table,),
                                 ((k, v) for k, v in self._metadata.items()
                                  if k in new_fields))

        # Exceptionally, the alter tables are forcefully committed
        self._conn.commit()

    def _get_field_transformer(self, field_name):
        try:
            # Fast scenario: the transformer is cached in its place
            return self._field_transformers[field_name]
        except KeyError:
            pass
        if self._metadata is None:
            self._cursor.execute("SELECT * FROM `%s_tamd`" % (self._table,))
            result = self._cursor.fetchall()
            self._metadata = {k: v for k, v in result}

        field_type = self._metadata.get(field_name)
        if field_type is None:
            tf = transformers.Identity
        elif field_type == "pickle":
            tf = transformers.PickleTransformer
        elif field_type == "json":
            tf = transformers.JsonTransformer
        elif field_type == "yaml":
            tf = transformers.YamlTransformer
        else:
            raise ValueError("Unknown field_type for TAD4BJ: %s" % field_type)

        self._field_transformers[field_name] = tf
        return tf

    def get_value(self, jobid, field, raw_return=False):
        self._cursor.execute("SELECT `%s` FROM `%s` WHERE id=?" %
                             (field, self._table), (jobid,))
        ret = self._cursor.fetchone()[0]

        if raw_return or ret is None:
            return ret

        # Transform the value before returning it
        return self._get_field_transformer(field).from_db(ret)

    def set_value(self, jobid, field, parameter, raw_parameter=False):
        if raw_parameter:
            value = parameter
        else:
            value = self._get_field_transformer(field).to_db(parameter)

        ex = self._cursor.execute("UPDATE `%s` SET `%s` = ? WHERE id = ?" %
                                  (self._table, field), (value, jobid))
        if ex.rowcount == 0:
            self._cursor.execute("INSERT INTO `%s` (id, `%s`) VALUES (?, ?)" %
                                 (self._table, field), (jobid, value))

    def set_values(self, jobid, fields, parameters, raw_parameters=False):
        set_str = ", ".join("`%s` = ?" % field_name for field_name in fields)

        if raw_parameters:
            values = list(parameters)
        else:
            values = list()

            for field, parameter in zip(fields, parameters):
                values.append(self._get_field_transformer(field).to_db(parameter))

        values.append(jobid)

        ex = self._cursor.execute("UPDATE `%s` SET %s WHERE id = ?" %
                                  (self._table, set_str), values)
        if ex.rowcount == 0:
            fields_str = ", ".join("`%s`" % field_name for field_name in fields)
            question_marks = ", ".join(["?"] * (len(fields) + 1))
            self._cursor.execute("INSERT INTO `%s` (%s, id) VALUES (%s)" %
                                 (self._table, fields_str, question_marks), values)

    def get_handler(self, jobid):
        return handlers.JobHandler(self, jobid)

    def __iter__(self):
        self._cursor.execute("SELECT `id` FROM %s" % (self._table,))
        return (res[0] for res in self._cursor.fetchall())

    def __contains__(self, item):
        self._cursor.execute("SELECT COUNT(*) FROM %s WHERE id = ?" %
                             (self._table,), (item,))
        return self._cursor.fetchone()[0] == 1

    def __getitem__(self, item):
        self._cursor.execute("SELECT * FROM %s WHERE id = ?" %
                             (self._table,), (item,))
        row_raw = self._cursor.fetchone()

        if row_raw is None:
            return KeyError("No row with id=%s" % item)

        row_namedtuple = self._get_row_namedtuple()
        processed_values = list()
        for field, value in zip(row_namedtuple._fields, row_raw):
            if value is not None:
                processed_values.append(
                    self._get_field_transformer(field).from_db(value)
                )
            else:
                processed_values.append(None)

        return row_namedtuple(*processed_values)

    def __len__(self):
        self._cursor.execute("SELECT COUNT(*) FROM %s" %
                             (self._table,))
        return self._cursor.fetchone()[0]


class DummyDataStorage(object):
    """Seems  DataStorage, but does nothing and doesn't raise expcetions (almost)."""

    def __init__(self, *args, **kwargs):
        pass

    def close(self):
        pass

    def clear(self, remove_tables=False):
        pass

    def prepare(self, schema):
        raise NotImplementedError("Refusing to dummy-prepare a table. I am a dummy.")

    def update(self, schema):
        pass

    def get_value(self, jobid, field, raw_return=False):
        return None

    def set_value(self, jobid, field, value, raw_parameter=False):
        pass

    def set_values(self, jobid, fields, values, raw_parameters=False):
        pass

    def get_handler(self, jobid=1):
        return handlers.JobHandler(self, jobid)
